/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Run
Version:   1.0
ID:        ID_RUN
Category:  Command
Date:      April 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2004.  All rights reserved.
Short:     The execute class runs native binary programs.
Keywords:  execute, executable, run, binary, program, task

###DESCRIPTION###
<p>The Run class allows you to launch binary files that are stored in the
file system.  Executable files must be native to the system that Athene is
running on.  For example, if you are running Athene on a Linux system, the
Run class will allow you to run X11 programs and popular applications such
as Netscape.</p>

<p>The Run class is not restricted to launching executable files.  For
instance, you can launch picture, text and custom file types.  The exact
behaviour that occurs when launching a data file is dependent on the
settings found in the user's &quot;user:config/associations.cfg&quot;
file, which holds the launch settings for each registered class and file
type.  In most cases a script will be run which can handle the file
data.</p>

<p>Typical usage of the Run class in a DML script is demonstrated in the
following examples:</p>

<pre>
   &lt;run src="netscape"/&gt;
   &lt;run mode="view" src="pictures:screenshot.png"/&gt;
</pre>

<p>If you need to send arguments to an executable file, use the
<a href="#Args">Args</a> field as in this example:</p>

<pre>
   &lt;run src="myprogram" args="arg1 arg2 arg3"/&gt;
</pre>

<p>Like all commands, you may make a run object static if you want it
to be activated in response to system events.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/data/config.h>
#include <pandora/modules/filesystem.h>
#include <pandora/modules/strings.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_RUN 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_RUN, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2004.  All rights reserved.",
   "August 2004",
   "Run"
};

static OBJECTPTR RunClass = 0;
static OBJECTPTR modFileSystem = 0;
static OBJECTPTR modStrings = 0;
static struct KernelBase *KernelBase = 0;
static struct FileSystemBase *FileSystemBase = 0;
static struct StringsBase *StringsBase = 0;

/*****************************************************************************
** Run class definition.
*/

struct Run {
   OBJECT_HEADER        /* Class header */
   LONG     Static;     /* Set if the Run object should reside in memory */
   OBJECTID OutputID;   /* Reference to an output object */
   LONG     Mode;       /* The type of action to perform in launching the file */
   LONG     Flags;
   LONG     ReturnCode;
   FLOAT    TimeOut;

   PRIVATE_FIELDS
   UBYTE  Buffer[1024];   /* The buffer is required in the structure to prevent forking problems */
   STRING Location;       /* The location of the program */
   STRING Args;           /* The arguments to pass to the program */
   MEMORYID LocationMID;
   MEMORYID ArgsMID;
   UBYTE Type[32];
};

#define RNL_OPEN  1
#define RNL_EDIT  2
#define RNL_PRINT 3
#define RNL_VIEW  4

#define RNF_FOREIGN    0x00000001
#define RNF_WAIT       0x00000002
#define RNF_RESETDIR   0x00000004
#define RNF_PRIVILEGED 0x00000008

static ERROR GET_Args(struct Run *, STRING *);
static ERROR GET_Location(struct Run *, STRING *);
static ERROR GET_Type(struct Run *, STRING *);

static ERROR SET_Args(struct Run *, STRING);
static ERROR SET_Location(struct Run *, STRING);
static ERROR SET_TimeOut(struct Run *, FLOAT *);
static ERROR SET_Type(struct Run *, STRING);

static struct FieldDef ModeLookup[] = {
   { "OPEN",  RNL_OPEN },
   { "EDIT",  RNL_EDIT },
   { "PRINT", RNL_PRINT },
   { "VIEW",  RNL_VIEW },
   { NULL, NULL }
};

static struct FieldDef RunFlags[] = {
   { "FOREIGN",    RNF_FOREIGN },
   { "WAIT",       RNF_WAIT },
   { "RESETDIR",   RNF_RESETDIR },
   { "PRIVILEGED", RNF_PRIVILEGED },
   { NULL, NULL }
};

static struct FieldArray RunFields[] = {
  { "Static",     0, FDF_LONG|FDF_RI,            0, NULL, NULL },
  { "Output",     0, FDF_OBJECTID|FDF_RI,        0, NULL, NULL },
  { "Mode",       0, FDF_LONG|FDF_LOOKUP|FDF_RW, (LONG)&ModeLookup, NULL, NULL },
  { "Flags",      0, FDF_LONGFLAGS|FDF_RI,       (LONG)&RunFlags, NULL, NULL },
  { "ReturnCode", 0, FDF_LONG|FDF_RI,            0, NULL, NULL },
  { "TimeOut",    0, FDF_FLOAT|FDF_RW,           0, NULL, SET_TimeOut },
  /*** Virtual fields ***/
  { "Args",     0, FDF_STRING|FDF_RW,   0, GET_Args,     SET_Args },
  { "Location", 0, FDF_STRING|FDF_RW,   0, GET_Location, SET_Location },
  { "Src",      0, FDF_STRING|FDF_RW,   0, GET_Location, SET_Location },
  { "Type",     0, FDF_STRING|FDF_RW,   0, GET_Type,     SET_Type },
  END_FIELD
};

static ERROR RUN_Activate(struct Run *, APTR);
static ERROR RUN_ClosingTag(struct Run *, APTR);
static ERROR RUN_Free(struct Run *, APTR);
static ERROR RUN_NewObject(struct Run *, APTR);
static ERROR RUN_ReleaseObject(struct Run *, APTR);
static ERROR RUN_Init(struct Run *, APTR Void);

static struct ActionArray RunActions[] = {
   { AC_Activate,      RUN_Activate },
   { AC_ClosingTag,    RUN_ClosingTag },
   { AC_Free,          RUN_Free },
   { AC_Init,          RUN_Init },
   { AC_NewObject,     RUN_NewObject },
   { AC_ReleaseObject, RUN_ReleaseObject },
   { NULL, NULL }
};

static void Print(struct Run *, STRING);
static ERROR GetClassCommand(struct Run *, OBJECTPTR, OBJECTPTR, CLASSID, STRING *);

/*****************************************************************************
** User Instructions.
*/

static BYTE RunUsage[] = {
"USAGE: RUN SRC=filename ARGS=\"arguments...\"\n\
\n\
Executable and data files can be run with this command.  For executable\n\
files, the file must be in a format that is native to the system that\n\
Athene is running on.  Arguments can be passed to executable programs\n\
via the Args string.  Data files can be in any format that is recognised\n\
by your \"user:config/associations.cfg\" file.\n\
\n\
Here are some examples:\n\
\n\
  run src=\"xterm\"\n\
  run src=\"demos:boing.dml\"\n\
  run src=\"pictures:screenshot.png\"\n\
  run src=\"bin:runscript\" args=\"-script templates:window.dml -wait\"\n"
};

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** File System module ***/

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   /*** Strings module ***/

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&RunClass, NULL,
      FID_BaseClassID|TLONG, ID_RUN,
      FID_Version|TFLOAT,    VER_RUN,
      FID_Name|TSTRING,      "Run",
      FID_Category|TLONG,    CCF_COMMAND,
      FID_Usage|TPTR,        RunUsage,
      FID_Actions|TPTR,      RunActions,
      FID_Fields|TPTR,       RunFields,
      FID_Size|TLONG,        sizeof(struct Run),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (modStrings)    { acFree(modStrings);    modStrings = 0; }
   if (modFileSystem) { acFree(modFileSystem); modFileSystem = 0; }
   if (RunClass)      { acFree(RunClass);      RunClass = 0; }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Activates a Run object.
###END###

*****************************************************************************/

static ERROR RUN_Activate(struct Run *Self, APTR Void)
{
   UBYTE buffer[800], commandfile[200];
   struct mtReadConfig read;
   struct ConfigEntry *entries;
   STRING location, args, command, str, mode;
   OBJECTPTR task, classes, config;
   LONG ClassID, SubClassID, i, j, amtentries, ext, flags, type;
   OBJECTID SystemClassesID;

   command = NULL;
   str = NULL;
   ClassID = 0;

   switch(Self->Mode) {
      case RNL_EDIT:  mode = "Edit"; break;
      case RNL_OPEN:  mode = "Open"; break;
      case RNL_PRINT: mode = "Print"; break;
      case RNL_VIEW:  mode = "View"; break;
      default:        mode = "Open";
   }

   /*** Get the location and argument strings ***/

   location = NULL;
   GET_Location(Self, &location);

   if (Self->Type[0]) {
      /* If the user has set a Type, there is no need to analyse the
      ** target file.  The type will either refer to a class name or
      ** a file association and we just need to determine what program
      ** to use for loading.
      */

      DPrintF("Run:","Predefined type: %s", Self->Type);

      if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
         FID_Location|TSTRING, "system:config/software/associations.cfg;user:config/associations.cfg",
         TAGEND) != ERR_Okay) {
         Print(Self, "Failed to load the \"user:config/associations.cfg\" file.");
         return(ERR_CreateObject);
      }

      read.Section = Self->Type;
      read.Item    = mode;
      if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) command = read.Data;

      if ((Self->Mode != RNL_OPEN) AND (!command)) {
         read.Section = Self->Type;
         read.Item    = "Open";
         if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) command = read.Data;
      }
   }
   else if (location) {
      /*** Check that the location refers to a valid file ***/

      if ((AnalyseLocation(location, &type) != ERR_Okay) OR (type != LOC_FILE)) {
         return(ERR_FileNotFound);
      }

      /*** Determine the start of the file extension, if there is one ***/

      for (ext=0; location[ext]; ext++);
      while ((ext > 0) AND (location[ext] != '.') AND (location[ext] != ':') AND (location[ext] != '/') AND (location[ext] != '\\')) ext--;
      if (location[ext] IS '.') ext++;
      else ext = 0;

      /* Determine the class that the file belongs to, or the association
      ** that should be used for executing the file.
      */

      for (i=0; location[i]; i++) if ((location[i] IS ':') OR (location[i] IS '/') OR (location[i] IS '\\')) break;
      if (!location[i]) ClassID = ID_TASK;
      else {
         /* Note that if the IdentifyFile() function fails it will be
         ** because the data does not belong to a specific class.
         ** The associations.cfg file will help us load files that
         ** do not have class associations later in this subroutine.
         */
         IdentifyFile(location, &ClassID, &SubClassID);
      }

      /*** If the file is an executable, run it immediately and return ***/

      if ((ClassID IS ID_TASK) OR (Self->Flags & RNF_FOREIGN)) {

         GET_Args(Self, &args);

         flags = NULL;

         if (Self->Flags & RNF_FOREIGN) flags |= TSF_FOREIGN;
         if (Self->Flags & RNF_WAIT) flags |= TSF_WAIT;
         if (Self->Flags & RNF_RESETDIR) flags |= TSF_RESETDIR;
         if (Self->Flags & RNF_PRIVILEGED) flags |= TSF_PRIVILEGED;

         if (CreateObject(ID_TASK, NF_CHILD, &task, NULL,
            FID_Location|TSTRING, location,
            FID_Output|TLONG,     Self->OutputID,
            FID_Args|TSTRING,     args,
            FID_Flags|TLONG,      flags,
            FID_TimeOut|TFLOAT,   Self->TimeOut,
            TAGEND) IS ERR_Okay) {

            if (acActivate(task) IS ERR_Okay) {
               if (Self->Flags & RNF_WAIT) {
                  GetFields(task, FID_ReturnCode|TLONG, &Self->ReturnCode,
                                  FID_TimeOut|TFLOAT,   &Self->TimeOut,
                                  TAGEND);

                  DPrintF("Run:","Return Code: %d", Self->ReturnCode);
               }

               acFree(task);
               if (!Self->TimeOut) return(ERR_TimeOut);
               else return(ERR_Okay);
            }
            else {
               acFree(task);
               Print(Self, "The executable program could not be loaded.");
               return(ERR_Activate);
            }
         }
         else {
            Print(Self, "Internal failure (A Task object could not be created).");
            return(ERR_CreateObject);
         }
      }

      /*** Load the associations configuration file ***/

      if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
         FID_Location|TSTRING, "system:config/software/associations.cfg;user:config/associations.cfg",
         TAGEND) != ERR_Okay) {
         Print(Self, "Failed to load the \"user:config/associations.cfg\" file.");
         return(ERR_CreateObject);
      }

      if (ClassID) {
         /* Get the name of the class and sub-class so that we can search
         ** the SystemAssociations object.
         */

         if (FastFindObject("SystemClasses", NULL, &SystemClassesID, 1, NULL) IS ERR_Okay) {
            if (AccessObject(SystemClassesID, 3000, &classes) IS ERR_Okay) {
               if ((!SubClassID) OR (GetClassCommand(Self, classes, config, SubClassID, &command) != ERR_Okay)) {
                  GetClassCommand(Self, classes, config, ClassID, &command);
               }
               ReleaseObject(classes);
            }
         }
      }

      /* Scan for customised file assocations.  These override the default
      ** class settings, so the user can come up with his own personal
      ** settings in circumstances where a class assocation is not suitable.
      */

      if (ext) {
         DPrintF("Activate:","[Run] Scanning for extension: %s", location+ext);

         if (GetFields(config, FID_AmtEntries|TLONG, &amtentries,
                               FID_Entries|TPTR,     &entries,
                               TAGEND) IS ERR_Okay) {
            for (i=0; (i < amtentries); i++) {
               if (StrCompare(entries[i].Item, "Extension", NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                  STRING extdata;
                  BYTE match;

                  extdata = entries[i].Data;
                  match = FALSE;
                  while ((*extdata) AND (match IS FALSE)) {
                     match = TRUE;
                     for (j=0; (location[ext+j]) AND (*extdata); j++, extdata++) {
                        if (location[ext+j] != *extdata) { match = FALSE; break; }
                     }
                     if ((location[ext+j]) OR ((*extdata != ';') AND (*extdata))) match = FALSE;

                     while ((*extdata) AND (*extdata != ';')) extdata++;
                     if (*extdata IS ';') extdata++;
                  }

                  if (match IS TRUE) {
                     ERROR error;

                     read.Section = entries[i].Section;
                     read.Item    = mode;
                     if ((error = Action(MT_ReadConfig, config, &read)) IS ERR_Okay) command = read.Data;

                     if ((Self->Mode != RNL_OPEN) AND ((error != ERR_Okay) OR (!command))) {
                        read.Section = Self->Type;
                        read.Item    = "Open";
                        if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) command = read.Data;
                     }
                     break;
                  }
               }
            }
         }
      }
   }
   else {
      DPrintF("@Activate:","[Run] Target file not specified.");
      return(ERR_FieldNotSet);
   }

   /* If no association exists for the file then use the default
   ** options if they are available.
   */

   if (!command) {
      read.Section = "default";
      read.Item = mode;
      if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) command = read.Data;
   }

   /*** Process the command string (must either be a binary or a script) ***/

   if (!command) {
      if (Self->OutputID) Print(Self, "Failed to identify the file's type.");
      else {
         OBJECTPTR dialog;

         if (Self->Type[0]) StrFormat(buffer, sizeof(buffer), "No programs registered to launch files of type \"%s\".", Self->Type);
         else if (ClassID) {
            OBJECTPTR class;
            STRING classname;
            if ((class = FindClass(ClassID, NULL))) {
               GetField(class, FID_Name, FT_STRING, &classname);
               StrFormat(buffer, sizeof(buffer), "File identified as belonging to class \"%s\", but no programs are registered against that class type.", classname);
            }
            else StrFormat(buffer, sizeof(buffer), "File identified as belonging to class %d, but no programs are registered against that class type.", ClassID);
         }
         else StrFormat(buffer, sizeof(buffer), "Failed to identify the type of file for location \"%s\".", Self->Location);

         if (CreateObject(ID_DIALOG, NF_UNTRACKED, &dialog, NULL,
            FID_Image|TSTRING,   "icons:items/error(48)",
            FID_Buttons|TSTRING, "NONE:Okay",
            FID_Title|TSTRING,   "File Launch Failure",
            FID_String|TSTRING,  buffer,
            TAGEND) IS ERR_Okay) {
            acShow(dialog);
         }
      }
      acFree(config);
      return(ERR_Failed);
   }

   /*** Extract the script or binary that we are asked to run ***/

   j = 0;
   commandfile[0] = 0;
   if (*command IS '"') {
      command++;
      while ((*command) AND (*command != '"') AND (j < sizeof(commandfile)-1)) commandfile[j++] = *command++;
      if (*command IS '"') command++;
   }
   else while ((*command) AND (*command != ' ') AND (j < sizeof(commandfile)-1)) commandfile[j++] = *command++;
   commandfile[j] = 0;

   while (*command IS ' ') command++;

   /*** Test the command reference to see if it is a script or a program and respond accordingly ***/

   ClassID = NULL;
   IdentifyFile(commandfile, &ClassID, NULL);

   if (ClassID IS ID_TASK) StrCopy(command, buffer, sizeof(buffer));
   else StrFormat(buffer, sizeof(buffer), "-target desktop -script \"%s\" -args %s", commandfile, command);

   /*** Free the config object as we no longer need it ***/

   acFree(config);

   /* Convert any references to {location} to the location of
   ** the file that we are attempting to run.
   */

   if (StrReplace(buffer, "{location}", location, &str, FALSE) IS ERR_Okay) {
      StrCopy(str, buffer, sizeof(buffer));
      FreeMemory(str);
   }

   /*** Launch the command ***/

   if (ClassID IS ID_TASK) {
      CreateObject(ID_TASK, NF_CHILD, &task, NULL,
         FID_Output|TLONG,     Self->OutputID,
         FID_Location|TSTRING, commandfile,
         FID_Args|TSTRING,     buffer,
         FID_Flags|TLONG,      TSF_RESETDIR,
         TAGEND);
   }
   else {
      CreateObject(ID_TASK, NF_CHILD, &task, NULL,
         FID_Location|TSTRING, "bin:runscript",
         FID_Output|TLONG,     Self->OutputID,
         FID_Args|TSTRING,     buffer,
         TAGEND);
   }

   if (!task) {
      Print(Self, "Internal failure (A Task object could not be created).");
      return(ERR_CreateObject);
   }

   if (acActivate(task) IS ERR_Okay) {
      acFree(task);
      return(ERR_Okay);
   }
   else {
      acFree(task);
      Print(Self, "The executable program could not be loaded.");
      return(ERR_Activate);
   }
}

/*****************************************************************************

###ACTION###
Name:  ClosingTag
Short: If the object is non-static, this action will start the execution process and then the object will self-destruct.
###END###

*****************************************************************************/

static ERROR RUN_ClosingTag(struct Run *Self, APTR Void)
{
   if (Self->Static IS FALSE) {
      acActivate(Self);
      acFree(Self);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Run: Free
*/

static ERROR RUN_Free(struct Run *Self, APTR Void)
{
   if (Self->Location)    { ReleaseMemory(Self->Location);   Self->Location = NULL; }
   if (Self->Args)        { ReleaseMemory(Self->Args);       Self->Args = NULL; }
   if (Self->LocationMID) { FreeMemoryID(Self->LocationMID); Self->LocationMID = NULL; }
   if (Self->ArgsMID)     { FreeMemoryID(Self->ArgsMID);     Self->ArgsMID = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Run: Init
*/

static ERROR RUN_Init(struct Run *Self, APTR Void)
{
   if (!Self->LocationMID) return(ObjectError(Self, ERH_Init, ERR_FieldNotSet));

   if (!Self->Mode) Self->Mode = RNL_OPEN;

   return(ERR_Okay);
}

/*****************************************************************************
** Run: NewObject
*/

static ERROR RUN_NewObject(struct Run *Self, APTR Void)
{
   Self->TimeOut    = 60 * 60 * 24;
   Self->ReturnCode = 0;
   return(ERR_Okay);
}

/*****************************************************************************
** Run: ReleaseObject
*/

static ERROR RUN_ReleaseObject(struct Run *Self, APTR Void)
{
   if (Self->Location) { ReleaseMemory(Self->Location); Self->Location = NULL; }
   if (Self->Args)     { ReleaseMemory(Self->Args);     Self->Args = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Args
Short:  Set this field to send the executable program a range of arguments.
Type:   STRING
Status: Read/Write

###DESCRIPTION###
<p>If the executable program accepts arguments, then you may send them to
the program by setting this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Args(struct Run *Self, STRING *Value)
{
   if (Self->Args) {
      *Value = Self->Args;
      return(ERR_Okay);
   }
   else if (!Self->ArgsMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->ArgsMID, MEM_READ, (void **)&Self->Args) IS ERR_Okay) {
      *Value = Self->Args;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_Args(struct Run *Self, STRING Value)
{
   LONG i;

   if (Self->Args)    { ReleaseMemory(Self->Args);   Self->Args = NULL; }
   if (Self->ArgsMID) { FreeMemoryID(Self->ArgsMID); Self->ArgsMID = NULL; }

   if ((Value) AND (*Value)) {
      if ((Value[0] IS '1') AND (!Value[1])) return(ERR_Okay); /* Ignore arguments of "1" */
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->Args, &Self->ArgsMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Args[i] = Value[i];
         Self->Args[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Defines special options to use when running programs.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Optional flags that affect the behaviour of the Run class may be specified in this field.</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">FOREIGN</td><td>If the program that you are running is not native to Athene or Pandora then you must set the FOREIGN flag so that the system does not attempt to treat it as a native program.</td></tr>
<tr><td width="10%">RESETDIR&nbsp;&nbsp;</td><td>Normally when a program is launched, it will start in the same directory as your own process.  If you would rather that the program starts in the same directory as its executable, set the RESETDIR flag.</td></tr>
<tr><td width="10%">WAIT</td><td>If this option is set then your task will be stop and wait for the other program to finish when launched.  This can be very useful when running batch scripts.  See the TimeOut field if you wish to limit the amount of time that you are prepared to wait.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:     Location
Synonyms: Src
Short:    Defines the location/name of the executable program.
Type:     STRING
Status:   Read/Write

###DESCRIPTION###
<p>It is compulsory that this field is set to the location of the
executable program that you wish to run.  Directory assigns are fully
supported, so you may use this file-system feature in specifying the
location.</p>

<p>If you are running Athene on a Unix system then you should know that the
PATH variable is supported, so you do not need to fully-qualify the location
if it is in the path.</p>

###END###

*****************************************************************************/

static ERROR GET_Location(struct Run *Self, STRING *Value)
{
   if (Self->Location) {
      *Value = Self->Location;
      return(ERR_Okay);
   }
   else if (!Self->LocationMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->LocationMID, MEM_READ, (void **)&Self->Location) IS ERR_Okay) {
      *Value = Self->Location;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_Location(struct Run *Self, STRING Value)
{
   STRING path;
   LONG i;

   if (Self->Location)    { ReleaseMemory(Self->Location);   Self->Location = NULL; }
   if (Self->LocationMID) { FreeMemoryID(Self->LocationMID); Self->LocationMID = NULL; }

   if ((Value) AND (*Value)) {
      if (ResolveAssign(Value, &path, RSF_APPROXIMATE) IS ERR_Okay) {
         for (i=0; path[i]; i++);
         if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->Location, &Self->LocationMID) IS ERR_Okay) {
            for (i=0; path[i]; i++) Self->Location[i] = path[i];
            Self->Location[i] = 0;
            FreeMemory(path);
            return(ERR_Okay);
         }
         else {
            FreeMemory(path);
            return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
         }
      }
      else {
         /* The location does not necessarily have to be a file - it could
         ** be a target like an http:// location for launching a web browser
         ** for example.
         */

         for (i=0; Value[i]; i++);
         if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->Location, &Self->LocationMID) IS ERR_Okay) {
            for (i=0; Value[i]; i++) Self->Location[i] = Value[i];
            Self->Location[i] = 0;
            return(ERR_Okay);
         }
         else {
            FreeMemory(path);
            return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
         }
      }
   }
   else return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Mode
Short:  The type of launch mode to use when launching data files.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>The Mode field determines the type of action that will be performed when
running a data file.  The default mode is 'Open', which will open the
data file in the preferred user mode (in most cases a viewing application
will be loaded for this purpose).  Other mode types that are available to
you are listed in the following table:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Value</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">EDIT</td><td>Edit mode is used for running an application that can edit the data file.</td></tr>
<tr><td width="10%">OPEN</td><td>Open mode will open the file in a way that is appropriate for the data in question (for instance, opening an audio file will play the audio data).</td></tr>
<tr><td width="10%">PRINT&nbsp;&nbsp;</td><td>Print mode is used to instantly print a data file to the default printer.  Printing can only be used on data files that have some form of visual representation.</td></tr>
<tr><td width="10%">VIEW</td><td>View mode is used to view the file data in the user display.  This mode can only be used on data files that have some form of visual representation.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ReturnCode
Short:  The code returned from the running process is stored here.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>If you need to know the return code of a process after it has finished
running, you can read the ReturnCode field.  You will need to specify the
WAIT flag as an option, as the return code can only be retrieved if you
wait for the process to finish.</p>

<p>If the return code value is unobtainable, this field returns a default
value of 0.  You can change the default return code by writing this field
prior to initialisation.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Static
Short:  Set to TRUE to make the object static.
Type:   BOOLEAN
Status: Read/Init

###DESCRIPTION###
<p>By default, a run object will activate itself and then self-destruct
when a closing tag is received.  If you would rather that the object stays in
the system, set this field to TRUE.  If you do this, the only way to get the
run object to perform is to call the Activate() action.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   TimeOut
Short:  Defines the maximum amount of time to wait for a process to return.
Type:   FLOAT
Status: Read/Set

###DESCRIPTION###
<p>The Run class gives you the option to wait on a process until it returns.
By default the code will wait indefinitely, but you can prevent this by
setting the TimeOut field to the number of seconds that you are prepared
to wait for the process to return.  If the TimeOut is reached, the ReturnCode
field will be set to ERR_TimeOut.  The process will be allowed to continue
running in the background if it does not complete within the timeout period.</p>

<p>The timeout value is measured in seconds and is cast as floating point so
that you can achieve sub-second precision.  By setting the TimeOut to any
value other than zero, you will automatically enable the WAIT flag option.</p>

###END###

*****************************************************************************/

static ERROR SET_TimeOut(struct Run *Self, FLOAT *Value)
{
   if (*Value < 0) Self->TimeOut = 0;
   else Self->TimeOut = *Value;

   if (Self->TimeOut <= 0) Self->Flags &= ~RNF_WAIT;
   else Self->Flags |= RNF_WAIT;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Type
Short:  The class or filetype association to use for running the target file.
Type:   STRING [32]
Status: Get/Set

###DESCRIPTION###
<p>By default the Run class will examine target files to determine their type
before launching the program that handles that data type.  This is not
always a viable method of execution however - trying to run a browser for a
file location such as http://localhost/ for instance will fail.  You may also
have good reasons for running a program associated with a particular filetype
but don't want to actually load any file into the program on startup.</p>

<p>To manage either instance you will need to set the Type field to the name of
the class or file association that handles the file type (for a list of
preconfigured file associations please check the
system:config/software/associations.cfg file).  In the aforementioned browser
example, the correct Type setting would be "HTML".</p>

###END###

*****************************************************************************/

static ERROR GET_Type(struct Run *Self, STRING *Value)
{
   *Value = Self->Type;
   return(ERR_Okay);
}

static ERROR SET_Type(struct Run *Self, STRING Value)
{
   WORD i;

   if ((!Value) OR (!*Value)) return(ERR_NoData);

   for (i=0; (Value[i]) AND (i < sizeof(Self->Type)-1); i++) Self->Type[i] = Value[i];
   Self->Type[i] = 0;
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: Print()
*/

static void Print(struct Run *Self, STRING Buffer)
{
   struct acDataChannel channel;

   if (Self->OutputID) {
      channel.ObjectID = Self->Head.UniqueID;
      channel.DataType = DATA_TEXT;
      channel.Version  = 1;
      channel.Buffer   = Buffer;
      for (channel.Size=0; Buffer[channel.Size]; channel.Size++);
      channel.TotalEntries = 1;
      ActionMsg(AC_DataChannel, Self->OutputID, &channel);
   }
   else DPrintF("Run:", Buffer);
}

/*****************************************************************************
** Internal: TestClass()
**
** Scan the SystemClasses object looking to extract the correct name for
** ClassID.  Then scan the Assocations object for an entry that is registered
** against the given class type.
*/

static ERROR GetClassCommand(struct Run *Self, OBJECTPTR Classes, OBJECTPTR Associations, CLASSID ClassID, STRING *Command)
{
   BYTE classname[32];
   struct mtReadConfig read;
   struct ConfigEntry *entries;
   LONG amtentries, i, j;

   if ((!ClassID) OR (!Classes) OR (!Command)) return(SendError(ERH_Function, ERR_Args));

   if (GetFields(Classes, FID_AmtEntries|TLONG, &amtentries,
                          FID_Entries|TPTR,     &entries,
                          TAGEND) IS ERR_Okay) {

      for (i=0; i < amtentries; i++) {
         if (StrCompare(entries[i].Item, "ClassID", NULL, STR_MATCHLENGTH) IS ERR_Okay) {
            if (ClassID IS StrToInt(entries[i].Data)) {
               for (j=0; (entries[i].Section[j]) AND (j < sizeof(classname)-1); j++) classname[j] = entries[i].Section[j];
               classname[j] = 0;

               if (GetFields(Associations, FID_AmtEntries|TLONG, &amtentries,
                                           FID_Entries|TPTR,     &entries,
                                           TAGEND) IS ERR_Okay) {

                  for (i=0; i < amtentries; i++) {
                     if (StrCompare(entries[i].Item, "Class", NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        if (StrCompare(entries[i].Data, classname, NULL, STR_MATCHLENGTH) IS ERR_Okay) {

                           read.Section = entries[i].Section;
                           if (Self->Mode IS RNL_EDIT)       read.Item = "Edit";
                           else if (Self->Mode IS RNL_OPEN)  read.Item = "Open";
                           else if (Self->Mode IS RNL_PRINT) read.Item = "Print";
                           else if (Self->Mode IS RNL_VIEW)  read.Item = "View";
                           else read.Item = "Open";

                           if (Action(MT_ReadConfig, Associations, &read) IS ERR_Okay) *Command = read.Data;
                           return(ERR_Okay);
                        }
                     }
                  }
               }
               else return(SendError(ERH_Function, ERR_GetField));

               return(ERR_Search);
            }
         }
      }
      return(ERR_Search);
   }
   else return(SendError(ERH_Function, ERR_GetField));
}
