/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Wait
Version:   1.0
Status:    Stable
ID:        ID_WAIT
Date:      October 2003
Category:  Command
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2002-2003.  All rights reserved.
Short:     The wait class is used to temporarily halt object script processing.
Keywords:  wait, pause, stop, time, halt

###DESCRIPTION###
<p>The Wait class provides you with the ability to halt script processing
temporarily.  It is most useful in situations where you need to wait for a
user response.  For example, if you pop up a confirmation dialog box during
batch script execution, you will need to wait for the user to respond before
continuing on with the rest of the script.  Failure to wait may cause multiple
dialog boxes to be raised and the script may end prematurely, causing problems
for the user.</p>

<p>The Wait class works by monitoring an object for activity while the task
goes to sleep.  When that object responds in the way that you expect, or if
the object is destroyed earlier than expected, the task will wake up and
processing will continue as normal.  The sleeping process occurs when a wait
object is activated or the ClosingTag action is used.  A wait object cannot
be made static as this defeats the purpose of the class.  If you need to
respond to an action during run-time operations, you should use the Action
class instead.</p>

<p>If a wait object is destroyed before the monitored object draws a response,
the sleep process will break and continue as if a response was gained.</p>

<p>An extra feature also allows you to use wait objects to sleep for limited
periods of time.  By setting the <a href="#Seconds">Seconds</a> field, a
wait object can put your process to sleep for the period of time that you
specify.  This field is a floating point unit, so you can also wait for
fractions of a second if you need precise scheduling.</p>

<p>Here are some DML examples demonstrating usage of the Wait class:</p>

<pre>
  &lt;wait object="[window]" action="free"/&gt;

  &lt;wait seconds="0.5"/&gt;
</pre>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/modules/strings.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_WAIT 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_WAIT, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2002.  All rights reserved.",
   "January 2002",
   "Wait"
};

static struct KernelBase  *KernelBase = 0;
static struct StringsBase *StringsBase = 0;
static OBJECTPTR StringsMod = 0;
static OBJECTPTR WaitClass = 0;
static LONG glBreakMessageID = 0;
static struct ActionTable *ActionTable = 0;

/*****************************************************************************
** Class definition.
*/

struct Wait {
   OBJECT_HEADER        /* Class header */
   OBJECTID ObjectID;
   LONG     Static;
   DOUBLE   Seconds;

   PRIVATE_FIELDS
   OBJECTID ClassID;      /* The class of the object in the ObjectID field */
   ACTIONID ActionID;     /* The related action or method ID */
   BYTE     Action[40];   /* The name of the action or method to monitor */
};

static ERROR GET_Action(struct Wait *, STRING *);
static ERROR GET_Custom(struct Wait *, STRING *);

static ERROR SET_Action(struct Wait *, STRING);
static ERROR SET_Custom(struct Wait *, STRING);
static ERROR SET_Object(struct Wait *, OBJECTID *);

static struct FieldArray WaitFields[] = {
   { "Object",  0, FDF_OBJECTID|FDF_RW,   0, NULL, SET_Object },
   { "Static",  0, FDF_LONG|FDF_RW,   0, NULL, NULL },
   { "Seconds", 0, FDF_DOUBLE|FDF_RW, 0, NULL, NULL },
   /*** Virtual fields ***/
   { "Action", 0, FDF_STRING|FDF_RW, 0, GET_Action, SET_Action },
   { "Custom", 0, FDF_STRING|FDF_RW, 0, GET_Custom, SET_Custom },
   /*** Synonyms ***/
   { "Method", 0, FDF_STRING|FDF_RW, 0, GET_Action, SET_Action },
   END_FIELD
};

static ERROR WAIT_ActionNotify(struct Wait *, struct acActionNotify *);
static ERROR WAIT_Activate(struct Wait *, APTR);
static ERROR WAIT_ClosingTag(struct Wait *, APTR);
static ERROR WAIT_Init(struct Wait *, APTR);

static struct ActionArray WaitActions[] = {
   { AC_ActionNotify, WAIT_ActionNotify },
   { AC_Activate,     WAIT_Activate },
   { AC_ClosingTag,   WAIT_ClosingTag },
   { AC_Init,         WAIT_Init },
   { NULL, NULL }
};

/*****************************************************************************
** User Instructions.
*/

static BYTE WaitUsage[] = {
   "This command class is not designed for CLI usage.\n"
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &StringsMod, NULL, FID_Name|TSTRING, "strings", TAGEND) IS ERR_Okay) {
      if (GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (!(ActionTable = ActionList())) {
      return(ObjectError(argModule, ERH_InitModule, ERR_Failed));
   }

   glBreakMessageID = AllocateID(IDTYPE_MESSAGE);

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&WaitClass, NULL,
      FID_BaseClassID|TLONG, ID_WAIT,
      FID_Version|TFLOAT,    VER_WAIT,
      FID_Name|TSTRING,      "Wait",
      FID_Category|TLONG,    CCF_COMMAND,
      FID_Usage|TPTR,        WaitUsage,
      FID_Actions|TPTR,      WaitActions,
      FID_Fields|TPTR,       WaitFields,
      FID_Size|TLONG,        sizeof(struct Wait),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (StringsMod) { Action(AC_Free, StringsMod, NULL); StringsMod  = NULL; }
   if (WaitClass)  { Action(AC_Free, WaitClass, NULL);  WaitClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Wait: ActionNotify
*/

static ERROR WAIT_ActionNotify(struct Wait *Self, struct acActionNotify *Args)
{
   /* If the action is customised, perform a check for the name of
   ** the custom action that we're waiting for.
   */

   if (Self->ActionID IS AC_Custom) {
      struct acCustom *custom;
      custom = (struct acCustom *)Args->Args;
      if (StrCompare(Self->Action, custom->String, 0, STR_MATCHLENGTH) != ERR_Okay) {
         /*** String doesn't match so return immediately ***/
         return(ERR_Okay);
      }
   }

   /*** Break our wait loop by placing a special message on the queue ***/

   DPrintF("ActionNotify:","[Wait] Action received - breaking out of sleep mode.");

   SendMessage(NULL, glBreakMessageID, NULL, NULL, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** Wait: Activate
*/

static ERROR WAIT_Activate(struct Wait *Self, APTR Void)
{
   struct Message message;
   struct NextMsg nextmsg;
   OBJECTPTR object;
   ERROR error;

   if (Self->Seconds > 0) {
      /*** Wait for a specified number of seconds, then return ***/
      WaitTime(Self->Seconds, (Self->Seconds - (LONG)Self->Seconds) * 1000000);
      return(ERR_Okay);
   }
   else if ((Self->ActionID) AND (Self->ObjectID)) {

      /* Subscribe to the object, then wait for ActionNotify to receive
      ** notification of the object's demise or execution.  Our ActionNotify
      ** will then send us a user message which this loop will capture and
      ** respond to.
      */

      if ((error = AccessObject(Self->ObjectID, 3000, &object)) IS ERR_Okay) {
         SubscribeAction(object, AC_Free, Self);
         if (Self->ActionID != AC_Free) SubscribeAction(object, Self->ActionID, Self);
         ReleaseObject(object);

         DPrintF("Activate()","[Wait] Entering sleep mode...");

         while (ProcessMessages(&nextmsg, MF_WAIT) IS ERR_Okay) {
            if (nextmsg.Count > 0) {
               if (GetMessage(NULL, NULL, NULL, &message, sizeof(struct Message)) IS ERR_Okay) {
                  if ((message.Type IS MSGID_QUIT) OR (message.Type IS glBreakMessageID)) break;
                  else DPrintF("@Wait:","Received unknown message type #%d.", message.Type);
               }
            }
         }

         DPrintF("Activate:","[Wait] Break message received - I'm awake!");

         return(ERR_Okay);
      }
      else {
         DPrintF("Activate()","[Wait] Could not access object #%d, error %d.", Self->ObjectID, error);
         return(ERR_AccessObject);
      }
   }
   else return(ObjectError(Self, ERH_Activate, ERR_FieldNotSet));
}

/*****************************************************************************
** Wait: ClosingTag
*/

static ERROR WAIT_ClosingTag(struct Wait *Self, APTR Void)
{
   if (Self->Static) return(ERR_Okay);

   Action(AC_Activate, Self, NULL);
   Action(AC_Free, Self, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** Wait: Init
*/

static ERROR WAIT_Init(struct Wait *Self, APTR Void)
{
   if ((!Self->ActionID) AND (Self->ObjectID)) {
      SET_Action(Self, Self->Action);
   }

   if ((!Self->ActionID) AND (Self->Seconds <= 0)) return(ObjectError(Self, ERH_Init, ERR_FieldNotSet));

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Action
Short:  Set this field to specify the action or method that you want to wait for.
Type:   STRING
Status: Read/Set

###DESCRIPTION###
<p>This field must be set to determine the action or method that will break
the sleeping state induced by a wait object.  A list of valid action names
are provided in the Action List Manual.</p>

###END###

*****************************************************************************/

static ERROR GET_Action(struct Wait *Self, STRING *Action)
{
   *Action = Self->Action;
   return(ERR_Okay);
}

static ERROR SET_Action(struct Wait *Self, STRING Value)
{
   struct MethodArray *MethodTable;
   OBJECTPTR class;
   LONG i, TotalMethods;

   Self->ActionID = NULL;

   if ((!Value) OR (!*Value)) return(ERR_Okay);

   for (i=0; (Value[i]) AND (i < sizeof(Self->Action)-1); i++) Self->Action[i] = Value[i];
   Self->Action[i] = 0;

   /*** Search for a matching action to get the action ID ***/

   for (i=1; ActionTable[i].Name; i++) {
      if (StrCompare(ActionTable[i].Name, Self->Action, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->ActionID = i;
         return(ERR_Okay);
      }
   }

   /*** Search for a method instead ***/

   if (Self->ObjectID) {
      if ((class = FindClass(Self->ClassID, NULL))) {
         if (GetFields(class, FID_Methods|TPTR,       &MethodTable,
                              FID_TotalMethods|TLONG, &TotalMethods,
                              TAGEND) IS ERR_Okay) {
            for (i=1; i < TotalMethods+1; i++) {
               if (StrCompare(Self->Action, MethodTable[i].Name, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                  Self->ActionID = MethodTable[i].MethodID;
                  return(ERR_Okay);
               }
            }
         }
      }
   }

   /*** Not a standard action or method, so assume that a custom action is being referenced ***/

   DPrintF("Wait:","[Wait:%d] Action \"%s\" unrecognised - assuming custom action.", Self->Head.UniqueID, Self->Action);

   Self->ActionID = AC_Custom;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Custom
Short:  Set this field if you need to wait on a customised action.
Type:   STRING
Status: Read/Set

###DESCRIPTION###
<p>If you need to wait on a customised action, set this field with the name
of the custom action instead of setting the Action field.</p>

###END###

*****************************************************************************/

static ERROR GET_Custom(struct Wait *Self, STRING *Value)
{
   *Value = Self->Action;
   return(ERR_Okay);
}

static ERROR SET_Custom(struct Wait *Self, STRING Value)
{
   LONG i;

   if ((!Value) OR (!*Value)) return(ERR_Okay);

   for (i=0; (Value[i]) AND (i < sizeof(Self->Action)-1); i++) Self->Action[i] = Value[i];
   Self->Action[i] = 0;

   Self->ActionID = AC_Custom;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Object
Short:  Refers to the object that you are waiting on.
Type:   OBJECTID
Status: Read/Set

###DESCRIPTION###
<p>The object that you want to monitor for activity will need to be specified
in this field.  You will also need to supply the name of the action that will
be monitored by setting the <a href="#Action">Action</a> field.</p>

###SEE ALSO###
Fields: Action

###END###

*****************************************************************************/

static ERROR SET_Object(struct Wait *Self, OBJECTID *Value)
{
   Self->ObjectID = *Value;
   Self->ClassID  = GetClassID(Self->ObjectID);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Seconds
Short:  If you want to wait for a limited amount of time, set the Seconds field.
Type:   DOUBLE
Status: Read/Init

###DESCRIPTION###
<p>The Seconds field allows you to use a wait object for short pauses.  By
setting this field to a value of 60 for example, you can wait for a period of
one minute.  Because this field is in floating point, you can also wait for
extremely small intervals (accurate to 1/1000000th of a second, subject to
CPU and system performance).</p>

###END###

*****************************************************************************/
